import autonetkit.ank as ank_utils
import autonetkit.log as log
import networkx as nx
import pprint
import itertools
import autonetkit.log as log

#TODO: need to handle case of switches along path.... won't work for just router connectivity path lengths
def allocate(G_phy, G_bgp):
    log.info("Allocating route reflectors")
    graph_phy = G_phy._graph
    for asn, devices in G_phy.groupby("asn").items():
        routers = [d for d in devices if d.is_router]
        router_ids = list(ank_utils.unwrap_nodes(routers))
        mapping_id_to_device = dict(zip(router_ids, routers)) # to reverse lookup id back to device

        subgraph_phy = graph_phy.subgraph(router_ids)
        if len(subgraph_phy) == 1:  
                continue # single node in graph, no ibgp

        betw_cen = nx.degree_centrality(subgraph_phy)

        ordered = sorted(subgraph_phy.nodes(), key = lambda x: betw_cen[x], reverse = True)

        rr_count = len(subgraph_phy)/4 or 1# Take top 20% to be route reflectors
        route_reflectors = ordered[:rr_count] # most connected x%
        log.debug("Chose route_reflectors %s" % route_reflectors)
        rr_clients = ordered[rr_count:] # the other routers
        route_reflectors = list(ank_utils.wrap_nodes(G_bgp, route_reflectors))
        rr_clients = list(ank_utils.wrap_nodes(G_bgp, rr_clients))

# distances (shortest path, physical graph) from rrs to clients
        path_lengths = {}
        for rr in route_reflectors:
            path = nx.single_source_shortest_path_length(subgraph_phy, rr)
            path_mapped = dict( (mapping_id_to_device[id], length) for (id, length) in path.items()) # ids to devices
            path_lengths[rr] = path_mapped

        G_bgp.update(route_reflectors, route_reflector = True) # mark as route reflector
        # rr <-> rr
        over_links = [(rr1, rr2) for rr1 in route_reflectors for rr2 in route_reflectors if rr1 != rr2] 
        G_bgp.add_edges_from(over_links, type = 'ibgp', direction = 'over')

        for client in rr_clients:
            ranked_rrs = sorted(route_reflectors, key = lambda rr: path_lengths[rr][client])
            parent_count = 2 # number of parents to connect to for each rr client
            parent_rrs = ranked_rrs[:parent_count]
            log.info("Route reflectors for %s are %s " % (client, parent_rrs))

            for parent in parent_rrs:
                # client -> rr
                #up_links = [(client, rr) for (client, rr) in itertools.product(rr_clients, route_reflectors)]
                G_bgp.add_edge(client, parent, type = 'ibgp', direction = 'up')
                # rr -> client
                #down_links = [(rr, client) for (client, rr) in up_links] # opposite of up
                G_bgp.add_edge(parent, client, type = 'ibgp', direction = 'down')

    log.debug("iBGP done")



import networkx as nx
import string
import os
import itertools
import math
import autonetkit.config
settings = autonetkit.config.settings
import autonetkit.log as log
import autonetkit.exception
#TODO: fallback to python if cStringIO version is not available
from cStringIO import StringIO
from collections import defaultdict

def load_graphml(input_data):
    #TODO: allow default properties to be passed in as dicts


    try:
        graph = nx.read_graphml(input_data)
    except IOError, e:
        acceptable_errors = set([
            2, # no such file or directory
            36, # input string too long for filename
            63, # input string too long for filename
            ])
        if e.errno in acceptable_errors:
            # try as data string rather than filename string
            try:
                input_pseduo_fh = StringIO(input_data) # load into filehandle to networkx
                graph = nx.read_graphml(input_pseduo_fh)
            except IOError:
                raise autonetkit.exception.AnkIncorrectFileFormat
            except IndexError:
                raise autonetkit.exception.AnkIncorrectFileFormat
        else:
            raise e

    # remove selfloops
    graph.remove_edges_from(edge for edge in graph.selfloop_edges())
#TODO: if selfloops then log that are removing

    letters_single = (c for c in string.lowercase) # a, b, c, ... z
    letters_double = ("%s%s" % (a, b) for (a, b) in itertools.product(string.lowercase, string.lowercase)) # aa, ab, ... zz
    letters = itertools.chain(letters_single, letters_double) # a, b, c, .. z, aa, ab, ac, ... zz
#TODO: need to get set of current labels, and only return if not in this set

    #TODO: add cloud, host, etc
    # prefixes for unlabelled devices, ie router -> r_a
    label_prefixes = {
            'router': 'r',
            'switch': 'sw',
            'server': 'se',
            }

    current_labels = set(graph.node[node].get("label") for node in graph.nodes_iter())
    unique_label = (letter for letter in letters if letter not in current_labels)

#TODO: make sure device label set
    ank_graph_defaults = settings['Graphml']['Graph Defaults']
    for key, val in ank_graph_defaults.items():
        if key not in graph.graph:
            graph.graph[key] = val

    # handle yEd exported booleans: if a boolean is set, then only the nodes marked true have the attribute. need to map the remainder to be false to allow ANK logic 
    #for node in graph.nodes(data=True):
        #print node

    all_labels = dict((n, d.get('label')) for n, d in graph.nodes(data=True))
    label_counts = defaultdict(list)
    for (node, label) in all_labels.items():
        label_counts[label].append(node)

    # set default name for blank labels to ensure unique
    try:
        blank_labels = [v for k, v in label_counts.items() if not k].pop() # strip outer list
    except IndexError:
        blank_labels = [] # no blank labels
    for index, node in enumerate(blank_labels):
        #TODO: log message that no label set, so setting default
        graph.node[node]['label'] = "none___%s" % index

    duplicates = [(k, v) for k, v in label_counts.items() if k and len(v) > 1]
    for label, nodes in duplicates:
        for node in nodes:
            #TODO: need to check they don't all have same ASN... if so then warn
            try:
                graph.node[node]['label'] = "%s_%s" % (graph.node[node]['label'], graph.node[node]['asn'])
            except KeyError:
                log.warning("Unable to set new label for duplicate node %s: %s" % (node,
                    graph.node[node].get("label")))

    boolean_attributes = set( k for n, d in graph.nodes(data=True)
            for k, v in d.items()
            if isinstance(v, bool)
            )

    for node in graph:
        for attr in boolean_attributes:
            if attr not in graph.node[node]:
                graph.node[node][attr] = False


#TODO: store these in config file
    ank_node_defaults = settings['Graphml']['Node Defaults']
    node_defaults = graph.graph['node_default'] # update with defaults from graphml
    for key, val in node_defaults.items():
        if val == "False":
            node_defaults[key] = False

    #TODO: do a dict update before applying so only need to iterate nodes once

    for key, val in ank_node_defaults.items():
        if key not in node_defaults or node_defaults[key] == "None":
            node_defaults[key] = val

    for node in graph:
        for key, val in node_defaults.items():
            if key not in graph.node[node]:
                graph.node[node][key] = val

    # map lat/lon from zoo to crude x/y approximation
    if graph.graph.get('Creator') == "Topology Zoo Toolset":
        all_lat = [graph.node[n].get('Latitude') for n in graph
                if graph.node[n].get("Latitude")]
        all_lon = [graph.node[n].get('Longitude') for n in graph
                if graph.node[n].get("Longitude")]

        lat_min = min(all_lat)
        lon_min = min(all_lon)
        lat_max = max(all_lat)
        lon_max = max(all_lon)
        lat_mean = (lat_max - lat_min) / 2
        lon_mean = (lon_max - lon_min) / 2
        lat_scale = 500/(lat_max - lat_min)
        lon_scale = 500/(lon_max - lon_min)
        for node in graph:
            lat = graph.node[node].get('Latitude') or lat_mean # set default to be mean of min/max
            lon = graph.node[node].get('Longitude') or lon_mean # set default to be mean of min/max
            graph.node[node]['y'] = -1 * lat * lat_scale
            graph.node[node]['x'] = lon * lon_scale

    if not( any(graph.node[n].get('x') for n in graph)
            and any(graph.node[n].get('y') for n in graph)):
# No x, y set, layout in a grid
        grid_length = int(math.ceil(math.sqrt(len(graph))))
        co_ords = [(x*100, y*100) for y in range(grid_length) for x in range(grid_length)]
        # (0,0), (100, 0), (200, 0), (0, 100), (100, 100) ....
        for node in sorted(graph):
            x, y = co_ords.pop(0)
            graph.node[node]['x'] = x
            graph.node[node]['y'] = y

    # and ensure asn is integer, x and y are floats
    for node in sorted(graph):
        graph.node[node]['asn'] = int(graph.node[node]['asn'])
        try:
            x = float(graph.node[node]['x'])
        except KeyError:
            x = 0
        graph.node[node]['x'] = x
        try:
            y = float(graph.node[node]['y'])
        except KeyError:
            y = 0
        graph.node[node]['y'] = y
        try:
            graph.node[node]['label']
        except KeyError:
            device_type = graph.node[node]['device_type']
            graph.node[node]['label'] = "%s_%s" % (label_prefixes[device_type], unique_label.next())

    ank_edge_defaults = settings['Graphml']['Edge Defaults']
    edge_defaults = graph.graph['edge_default']
    for key, val in ank_edge_defaults.items():
        if key not in edge_defaults or edge_defaults[key] == "None":
            edge_defaults[key] = val

    for src, dst in graph.edges():
        for key, val in edge_defaults.items():
            if key not in graph[src][dst]:
                graph[src][dst][key] = val

    # allocate edge_ids
    for src, dst in graph.edges():
        graph[src][dst]['edge_id'] = "%s_%s" % (graph.node[src]['label'], graph.node[dst]['label'])

# apply defaults
# relabel nodes
#other handling... split this into seperate module!
# relabel based on label: assume unique by now!
    #if graph.graph.get("Network") == "European NRENs":
        #TODO: test if non-unique labels, if so then warn and proceed with this logic
        # we need to map node ids to contain network to ensure unique labels
        #mapping = dict( (n, "%s__%s" % (d['label'], d['asn'])) for n, d in graph.nodes(data=True))

    mapping = dict( (n, d['label']) for n, d in graph.nodes(data=True))
    if not all( key == val for key, val in mapping.items()):
        nx.relabel_nodes(graph, mapping, copy=False) # Networkx wipes data if remap with same labels

    return graph



"""
AutoNetkit Utilities
"""

#from anm import overlay_node, overlay_edge
import autonetkit

def unwrap_nodes(nodes):
    """Unwrap nodes"""
    try:
        return nodes.node_id # treat as single node
    except AttributeError:
        return (node.node_id for node in nodes) # treat as list

def unwrap_edges(edges):
    """Unwrap edges"""
    return ( (edge.src_id, edge.dst_id) for edge in edges)

def unwrap_graph(overlay_graph):
    """Unwrap graph"""
    return overlay_graph._graph


def alphabetical_sort( l ): 
    """From http://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python"""
#TODO: fix as currently only handles strings - not objects with repr?
    import re 
    """ Sort the given iterable in the way that humans expect.""" 
    convert = lambda text: int(text) if text.isdigit() else text 
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ] 
    return sorted(l, key = alphanum_key)

def wrap_nodes(overlay_graph, nodes):
    """ wraps node id into node overlay """
#TODO: remove duplicate of this in ank.py
    return ( autonetkit.anm.overlay_node(overlay_graph._anm, overlay_graph._overlay_id, node)
            for node in nodes)



import time
import json
import itertools
import math
import autonetkit.ank as ank_utils
import autonetkit.log as log
import autonetkit.ank_json
import networkx as nx
from collections import defaultdict
import netaddr
import functools

messaging = autonetkit.ank_messaging.AnkMessaging()

#TODO: allow slack in allocations: both for ASN (group level), and for collision domains to allow new nodes to be easily added

try:
    import cPickle as pickle
except ImportError:
    import pickle

def subnet_size(host_count):
    """Returns subnet size"""
    import math
    host_count += 2 # network and broadcast
    return int(math.ceil(math.log(host_count, 2)))


@functools.total_ordering
class TreeNode(object):
    def __init__(self, graph, node):
        object.__setattr__(self, 'graph', graph)
        object.__setattr__(self, 'node', node)

#TODO: make thise fixed attributes, as only certain number needed here
    def __getattr__(self, attr):
        return self.graph.node[self.node].get(attr)

    def __setattr__(self, key, val):
        self.graph.node[self.node][key] = val

    def __lt__(self, other):
        if self.host and other.host:
            return self.host < other.host
        return self.node < other.node


#TODO: restore function that truncated subnets

    def __repr__(self):
        if self.host:
            return "%s %s" % (self.subnet, self.host)
        if self.loopback_group:
            return "Lo Gr %s: %s" % (self.group_attr, self.subnet)
        if self.group_attr:
            return "%s: %s" % (self.group_attr, self.subnet)
        if self.subnet:
            return "%s" % self.subnet
        return "TreeNode: %s" % self.node 

    def is_collision_domain(self):
        return (self.host and self.host.collision_domain)

    def is_loopback_group(self):
        return self.loopback_group

    def is_host(self):
        return self.host

    def children(self):
        return [TreeNode(self.graph, child) for child in self.graph.successors(self.node)]

class IpTree(object):
    def __init__(self, root_ip_block):
        self.unallocated_nodes = []
        self.graph = nx.DiGraph()
        self.root_node = None
        self.timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())
#taken_nodes -> these are nodes manually specified, eg in graphml
        self.node_id_counter = (i for i in itertools.count(0) if i not in self.graph)
        self.root_ip_block = root_ip_block

    def __len__(self):
        return len(self.graph)

    def __iter__(self):
        return iter(TreeNode(self.graph, node) for node in self.graph)

    @property
    def next_node_id(self):
        return self.node_id_counter.next()

    def add_parent_nodes(self, subgraph, level_counts):
        for level in range(32, 0, -1):
            try:
                current_count = float(level_counts[level]) # float so do floating point division
            except KeyError:
                continue # key not present - likely higher up tree
            parent_count = int(math.ceil(current_count/2))
            parent_level = level - 1
            level_counts[parent_level] += parent_count
            subgraph.add_nodes_from((self.next_node_id, {'prefixlen': parent_level}) for n in range(parent_count))

            if level_counts[parent_level] == 1:
                if parent_level == min(level_counts.keys()):
                    break # Reached top of tree

    def build_tree(self, subgraph, level_counts, nodes_by_level):
        smallest_prefix = min(level_counts.keys())
        for prefixlen in range(smallest_prefix, 32):
            #TODO: fix sorting here
            unallocated_children = set(nodes_by_level[prefixlen + 1])
            for node in sorted(nodes_by_level[prefixlen]):
                is_not_subnet = not ("host" in subgraph.node[node] or "group_attr" in subgraph.node[node])
                if is_not_subnet:
                    child_a = unallocated_children.pop()
                    subgraph.add_edge(node, child_a)
                    try:
                        child_b = unallocated_children.pop()
                        subgraph.add_edge(node, child_b)
                    except KeyError:
                        pass # single child, just attach

        root_node = nodes_by_level[smallest_prefix][0]
        return root_node

    def save(self):
        import os
        import gzip
        archive_dir = os.path.join("versions", "ip")
        if not os.path.isdir(archive_dir):
            os.makedirs(archive_dir)

        self.graph.graph['timestamp'] = self.timestamp
        data = autonetkit.ank_json.ank_json_dumps(self.graph)
#TODO: should this use the ank_json.jsonify_nidb() ?
        json_file = "ip_%s.json.gz" % self.timestamp
        json_path = os.path.join(archive_dir, json_file)
        log.debug("Saving to %s" % json_path)
        #with gzip.open(json_path, "wb") as json_fh:
        with open(json_path, "wb") as json_fh:
            json_fh.write(data)

    def build(self, group_attr = 'asn'):
        """Builds tree from unallocated_nodes,
        groupby is the attribute to build subtrees from"""
        subgraphs = []

# if network final octet is .0 eg 10.0.0.0 or 192.168.0.0, then add extra "dummy" node, so don't have a loopback of 10.0.0.0
# Change strategy: if just hosts (ie loopbacks), then allocate as a large collision domain
        if not len(self.unallocated_nodes):
# no nodes to allocate - eg could be no collision domains
            return

        unallocated_nodes = self.unallocated_nodes
        unallocated_nodes = sorted(unallocated_nodes, key = lambda x: x.get(group_attr))
        for attr_value, items in itertools.groupby(unallocated_nodes, key = lambda x: x.get(group_attr)):
# make subtree for each attr
            items = list(items)
            subgraph = nx.DiGraph()
            if all(item.is_l3device for item in items): 
                # Note: only l3 devices are added for loopbacks: cds allocate to edges not devices (for now) - will be fixed when move to proper interface model 
                parent_id = self.next_node_id
                prefixlen = 32 - subnet_size(len(items)) # group all loopbacks into single subnet
                subgraph.add_node(parent_id, prefixlen = prefixlen, loopback_group = True)
                for item in items:
                #subgraph.add_edge(node, child_a)
                    item_id = self.next_node_id
                    subgraph.add_node(item_id, prefixlen = 32, host = item)
                    subgraph.add_edge(parent_id, item_id)

                root_node = parent_id
                subgraphs.append(subgraph)
                subgraph.graph['root'] = root_node
                subgraph.node[root_node]['group_attr'] = attr_value
                continue # finished for loopbacks, continue only for collision domains

            for item in items:
                if item.collision_domain:
                    subgraph.add_node(self.next_node_id, prefixlen = 32 - subnet_size(item.degree()), host = item)
                if item.is_l3device:
                    subgraph.add_node(self.next_node_id, prefixlen = 32, host = item)

            # now group by levels
            level_counts = defaultdict(int)

            nodes_by_level = defaultdict(list)
            for node in subgraph.nodes():
                prefixlen = subgraph.node[node]['prefixlen']
                nodes_by_level[prefixlen].append(node)

            for level, nodes in nodes_by_level.items():
                level_counts[level] = len(nodes)

            self.add_parent_nodes(subgraph, level_counts)

            # rebuild with parent nodes
            nodes_by_level = defaultdict(list)
            for node in subgraph.nodes():
                prefixlen = subgraph.node[node]['prefixlen']
                nodes_by_level[prefixlen].append(node)

            root_node = self.build_tree(subgraph, level_counts, nodes_by_level)
            subgraphs.append(subgraph)

            subgraph.graph['root'] = root_node
            subgraph.node[root_node]['group_attr'] = attr_value

        global_graph = nx.DiGraph()
        subgraphs = sorted(subgraphs, key = lambda x: subgraph.node[subgraph.graph['root']]['group_attr'])
        root_nodes = [subgraph.graph['root'] for subgraph in subgraphs]
        root_nodes = []
        for subgraph in subgraphs:
            root_node = subgraph.graph['root']
            root_nodes.append(root_node)
            global_graph.add_node(root_node, subgraph.node[root_node])

        nodes_by_level = defaultdict(list)
        for node in root_nodes:
            prefixlen = global_graph.node[node]['prefixlen']
            nodes_by_level[prefixlen].append(node)

        level_counts = defaultdict(int)
        for level, nodes in nodes_by_level.items():
            level_counts[level] = len(nodes)

        self.add_parent_nodes(global_graph, level_counts)

# rebuild nodes by level
#TODO: make this a function
        nodes_by_level = defaultdict(list)
        for node in global_graph:
            prefixlen = global_graph.node[node]['prefixlen']
            nodes_by_level[prefixlen].append(node)

        global_root = self.build_tree(global_graph, level_counts, nodes_by_level)
        global_root = TreeNode(global_graph, global_root)

        for subgraph in subgraphs:
            global_graph = nx.compose(global_graph, subgraph)

        # now allocate the IPs
        global_prefix_len = global_root.prefixlen
        global_ip_block = netaddr.IPNetwork("%s/%s" % (self.root_ip_block, global_prefix_len))
        self.graph = global_graph

# add children of collision domains
        cd_nodes = [n for n in self if n.is_collision_domain()]
        for cd in cd_nodes:
            for edge in sorted(cd.host.edges()):
                #TODO: sort these
                child_id = self.next_node_id
                cd_id = cd.node
                global_graph.add_node(child_id, prefixlen = 32, host = edge)
                global_graph.add_edge(cd_id, child_id) # cd -> neigh (cd is parent)

#TODO: make allocate seperate step
        def allocate(node):
            #children = graph.successors(node)
            children = sorted(node.children())
            prefixlen = node.prefixlen
            subnet = node.subnet.subnet(prefixlen+1)

            if node.is_loopback_group(): # special case of single AS -> root is loopback_group
                #TODO: generalise this rather than repeated code with below
                #node.subnet = subnet.next() # Note: don't break into smaller subnets if single-AS
                iterhosts = node.subnet.iter_hosts() # ensures start at .1 rather than .0
                sub_children = node.children()
                for sub_child in sub_children:
                    sub_child.subnet = iterhosts.next()

                return

            for child in children:
                if child.is_collision_domain():
                    child.subnet = subnet.next()
                    iterhosts = child.subnet.iter_hosts() # ensures start at .1 rather than .0
                    sub_children = child.children()
                    for sub_child in sub_children:
                        sub_child.subnet = iterhosts.next()
                        log.debug( "Allocate sub_child to %s %s" % ( sub_child, sub_child.subnet))
                elif child.is_host():
                    child.subnet = subnet.next()
                elif child.is_loopback_group():
                    child.subnet = subnet.next()
                    iterhosts = child.subnet.iter_hosts() # ensures start at .1 rather than .0
                    sub_children = child.children()
                    for sub_child in sub_children:
                        sub_child.subnet = iterhosts.next()
                else:
                    child.subnet = subnet.next()
                    allocate(child) # continue down the tree

        global_root.subnet = global_ip_block
#TODO: fix this workaround where referring to the wrong graph
        global_root_id = global_root.node
        global_root = TreeNode(global_graph, global_root_id)
        allocate(global_root)

# check for parentless nodes

        self.graph = global_graph
        self.root_node = global_root

    def group_allocations(self):
        allocs = {}
        for node in self:
            if node.group_attr:
                #TODO: Also need to store the type
                allocs[node.group_attr] = [node.subnet]

        return allocs

    def add_nodes(self, nodes):
        self.unallocated_nodes += list(nodes)

    def walk(self):
        def list_successors(node):
            successors = self.graph.successors(node)
            if successors:
                children = [list_successors(n) for n in successors]
                return {node: children}
            return node

        return list_successors(self.root_node)

    def json(self):
        def list_successors(node):
            children = node.children()
            if children:
                children = [list_successors(n) for n in children]
                return {"name": node,
                        "subnet": node.subnet,
                        "children": children}
            return {"name": node, "subnet": node.subnet}

        if not self.root_node:
            log.debug("No root node set")
            return {"name": str(self.root_ip_block),
                    "subnet": str(self.root_ip_block),
                    "children": []}
            return
        return list_successors(self.root_node)

    
    def assign(self):
# assigns allocated addresses back to hosts
        edges = [n for n in self if n.is_host() and n.host.src]
        for edge in edges:
            #print "edge subnet", edge.subnet
            edge.host.ip_address = edge.subnet


#TODO: do we need to store loopback groups into advertise addresses?

        #loopback_groups = [n for n in self if n.is_loopback_group()]
        #for loopback_group in loopback_groups:
            #print loopback_group
    
        # don't look at host nodes now - use loopback_groups
        host_tree_nodes = [n for n in self if n.is_host() and n.host.is_l3device]
        #for host_tree_node in host_tree_nodes:
            #print host_tree_node, host_tree_node.subnet
        for host_tree_node in host_tree_nodes:
            host_tree_node.host.loopback = host_tree_node.subnet

        cds = [n for n in self if n.is_collision_domain()]
        for cd in cds:
            cd.host.subnet = cd.subnet

def assign_asn_to_interasn_cds(G_ip):
    G_phy = G_ip.overlay("phy")
    for collision_domain in G_ip.nodes("collision_domain"):
        neigh_asn = list(ank_utils.neigh_attr(G_ip, collision_domain, "asn", G_phy)) #asn of neighbors
        if len(set(neigh_asn)) == 1:
            asn = set(neigh_asn).pop() # asn of any neigh, as all same
        else:
            asn = ank_utils.most_frequent(neigh_asn) # allocate cd to asn with most neighbors in it
        collision_domain.asn = asn

    return

def allocate_ips(G_ip, infrastructure = True):
    """Can disable infrastructure, eg for ipv6, still want to alloc ipv4 loopbacks for router ids"""
    log.info("Allocating Host loopback IPs")
    ip_tree = IpTree("10.0.0.0")
    ip_tree.add_nodes(G_ip.nodes("is_l3device"))
    ip_tree.build()
    loopback_tree = ip_tree.json()
   # json.dumps(ip_tree.json(), cls=autonetkit.ank_json.AnkEncoder, indent = 4)
    #body = json.dumps({"ip_allocations": jsontree})
    #messaging.publish_compressed("www", "client", body)
    ip_tree.assign()
    G_ip.data.loopback_blocks = ip_tree.group_allocations()

    log.info("Allocating Collision Domain IPs")

    if infrastructure:
        ip_tree = IpTree("192.168.1.0")
        assign_asn_to_interasn_cds(G_ip)
        ip_tree.add_nodes(G_ip.nodes("collision_domain"))
        ip_tree.build()
        cd_tree = ip_tree.json()
        ip_tree.assign()
    else:
        cd_tree = {}

    total_tree = {
            'name': "ip",
            'children': 
                [loopback_tree, cd_tree],
                #[loopback_tree],
            }
    jsontree = json.dumps(total_tree, cls=autonetkit.ank_json.AnkEncoder, indent = 4)

    body = json.dumps({"ip_allocations": jsontree})
    messaging.publish_compressed("www", "client", body)

#TODO: need to update with loopbacks if wish to advertise also - or subdivide blocks?
    G_ip.data.infra_blocks = ip_tree.group_allocations()

    #ip_tree.save()



import autonetkit.log as log
import time
import re
import autonetkit.config as config
import autonetkit.ank_messaging as ank_messaging

try:
    import Exscript
except ImportError:
    log.warning("Deployment requires Exscript: "
    "pip install https://github.com/knipknap/exscript/tarball/master")

def deploy(host, username, dst_folder):
    tar_file = package(dst_folder)
    transfer(host, username, tar_file)
    extract(host, username, tar_file, dst_folder)

def package(src_dir, target = "netkit_lab"):
    log.info("Packaging %s" % src_dir)
    import tarfile
    import os
    tar_filename = "%s.tar.gz" % target
    tar = tarfile.open(os.path.join(tar_filename), "w:gz")
    tar.add(src_dir)
    tar.close()
    return tar_filename

def transfer(host, username, local, remote = None, key_filename = None):
    log.debug("Transferring lab to %s" % host)
    log.info("Transferring Netkit lab")
    if not remote:
        remote = local # same filename
    import paramiko
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy( paramiko.AutoAddPolicy())
    if key_filename:
        log.debug("Connecting to %s with %s and key %s" % (host, username, key_filename))
        ssh.connect(host, username = username, key_filename = key_filename)
    else:
        log.info("Connecting to %s with %s" % (host, username))
        ssh.connect(host, username = username)
    log.debug("Opening SSH for SFTP")
    ftp = ssh.open_sftp()
    log.debug("Putting file %s to %s" % (local, remote))
    ftp.put(local, remote)
    log.debug("Put file %s to %s" % (local, remote))
    ftp.close()

def extract(host, username, tar_file, cd_dir, timeout = 30, key_filename = None, verbosity = 0):
    """Extract and start lab"""
    log.debug("Extracting and starting lab on %s" % (host))
    log.info("Extracting and starting Netkit lab")
    from Exscript import Account
    from Exscript.util.start import start
    from Exscript.util.match import first_match
    from Exscript import PrivateKey
    from Exscript.protocols.Exception import InvalidCommandException

    messaging = ank_messaging.AnkMessaging()

    def starting_host(protocol, index, data):
        m = re.search('\\"(\S+)\\"', data.group(index))
        if m:
            hostname = m.group(1)
            log.info(data.group(index)) #TODO: use regex to strip out just the machine name
            body = {"starting": hostname}
            messaging.publish_json(body)

    def lab_started(protocol, index, data):
        log.info("Lab started on %s" % host)
        body = {"lab started": host}
        messaging.publish_json(body)

    def make_not_found(protocol, index, data):
        log.warning("Make not installed on remote host %s. Please install make and retry." % host)
        return

    def start_lab(thread, host, conn):
        conn.set_timeout(timeout)
        conn.add_monitor(r'Starting (\S+)', starting_host)
        conn.add_monitor(r'The lab has been started', lab_started)
        conn.add_monitor(r'make: not found', make_not_found)
        #conn.data_received_event.connect(data_received)
        conn.execute('cd %s' % cd_dir)
        conn.execute('lcrash -k')
        conn.execute("lclean")
        conn.execute('cd') # back to home directory tar file copied to
        conn.execute('tar -xzf %s' % tar_file)
        conn.execute('cd %s' % cd_dir)
        conn.execute('vlist')
        conn.execute("lclean")
        log.info("Starting lab")
        start_command = 'lstart -p5 -o --con0=none'
        try:
            conn.execute(start_command)
        except InvalidCommandException, error:
            if "already running" in str(error):
                time.sleep(1)
                conn.execute(start_command)
        first_match(conn, r'^The lab has been started')
        conn.send("exit")

    if key_filename:
        key = PrivateKey.from_file(key_filename)
        log.debug("Connecting to %s with username %s and key %s" % (host, username, key_filename))
        accounts = [Account(username, key = key)] 
    else:
        log.debug("Connecting to %s with username %s" % (host, username))
        accounts = [Account(username)] 

    hosts = ['ssh://%s' % host]
    start(accounts, hosts, start_lab, verbose = verbosity)
